# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import multiprocessing
import os
import time
from functools import lru_cache

import fastapi
import fastapi.responses
import liquid
from fastapi import FastAPI, Request, Depends
from sqlalchemy.orm import Session

from tos_execute import access_log, error_log
from tos_execute import config
from tos_execute import errors
from tos_execute import init
from tos_execute import models_activeworkflow as awmodels
from tos_execute import models_internal as intmodels
from tos_execute import queue
from tos_execute import store
from tos_execute import utils

app = FastAPI()


@lru_cache
def get_settings():
    return config.BasicSettings()


@app.on_event("startup")
async def startup_event():
    # initialize the jobs into the database
    init.init_job_map(config.BasicSettings())

    # write the process id to a file
    os.makedirs(f"/tmp/uvicorn_processes", exist_ok=True)
    with open(f"/tmp/uvicorn_processes/{os.getpid()}", "w+") as writefile:
        writefile.write("True")

    # give the other processes some time to start as well
    time.sleep(5)

    # only the "first" process will do job restarting
    all_procs = [int(f) for f in os.listdir(f"/tmp/uvicorn_processes")]
    if os.getpid() == min(all_procs):
        error_log.info(
            f"Process ID [{os.getpid()}] is restarting broken and queued jobs."
        )
        # init possibly queued jobs
        init.restart_queued_jobs()

    else:
        error_log.info(f"Process ID [{os.getpid()}] stands by.")

    # give the other processes some time to reach here as well, delete file afterwards
    time.sleep(5)
    os.unlink(f"/tmp/uvicorn_processes/{os.getpid()}")


@app.get("/appinfo")
def appinfo(settings: config.BasicSettings = Depends(get_settings)):
    """Prints the configuration of this instance."""
    return settings.dict()


@app.get("/joblist")
def joblist(
        db: Session = Depends(store.get_db),
        settings: config.BasicSettings = Depends(get_settings),
):
    """Prints a list of all found and configured TOS jobs."""

    init.init_job_map(settings)
    jobs = store.get_jobs_by_availability(True, db)

    return jobs


@app.post(
    "/tosexecutor",
    response_model=awmodels.ResponseCommon,
    response_model_exclude_none=True,
)
def tos_executor(
        payload: awmodels.RequestCommon,
        db: Session = Depends(store.get_db),
        settings: config.BasicSettings = Depends(get_settings),
):
    """
    Implements an remote agent API endpoint as specified by active_workflow,
    see [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md)
    for more details.
    """

    # Replace credentials in options
    payload.params.options = utils.replace_credentials(data=payload.params.options,
                                                       credentials=payload.params.credentials)

    if payload.method is not None and payload.method == "register":
        return register(
            payload=awmodels.RequestRegister.parse_obj(payload.dict()), db=db
        )

    elif payload.method is not None and payload.method == "receive":
        return receive(
            payload=awmodels.RequestReceive.parse_obj(payload.dict()),
            db=db,
            settings=settings,
        )

    elif payload.method is not None and payload.method == "check":
        return check(
            payload=awmodels.RequestCheck.parse_obj(payload.dict()),
            db=db,
            settings=settings,
        )

    raise errors.ActiveWorkflowException(
        name="Method Error",
        message=f"You did not supply a valid method. Valid methods are: register, receive, check",
    )


@app.post(
    "/tosexecutor/register",
    response_model=awmodels.ResponseRegister,
    response_model_exclude_none=True,
    status_code=fastapi.status.HTTP_200_OK,
)
def register(payload: awmodels.RequestRegister, db: Session = Depends(store.get_db)):
    """Register according to the [ActiveWorkflow Remote Agent API](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#register-method).
    Returns the description as well as the name and default options to the caller,
    allowing ActiveWorkflow to register this agent into the engine.
    """

    access_log.info(
        f"{datetime.datetime.now().isoformat()} - Register: {payload.dict()}"
    )

    response = awmodels.ResponseRegister()
    jobs = store.get_jobs_by_availability(True, db)
    available_jobs = f"#### Joblist\n\n"
    for job in jobs:
        available_jobs += f"* Name: `{job.name}`, Version: `{job.version}`\n"

    response.result.description = response.result.description.replace(
        "#### Joblist", available_jobs
    )

    return response


@app.post(
    "/tosexecutor/receive",
    response_model=awmodels.ResponseReceive,
    response_model_exclude_none=True,
)
def receive(
        payload: awmodels.RequestReceive,
        db: Session = Depends(store.get_db),
        settings: config.BasicSettings = Depends(get_settings),
):
    """Receives a message that will start a new TOS Job in the background.

    Handle message receive. Receive will start a new TOS Job as a BackgroundTask that will run in the background,
    in which case this endpoint will return a HTTP 202 Accepted message. By setting key-value pairs to the `context`
    key of a message, you can override the context set in the options part.
    Each receive has to contain a `job_name` and a `job_version` in the options.
    """

    access_log.info(
        f"{datetime.datetime.now().isoformat()} - Receive: {payload.dict()}"
    )

    if not utils.is_authorized(payload.params, settings):
        raise errors.UnauthorizedException(
            name="Unauthorized",
            message='Please supply valid credentials using the `api_key` option parameter referring to a stored credential key using Liquid templating: `"api_key": "{% credential credential_name %}"`.',
        )

    response = awmodels.ResponseReceive()
    response.result.memory = payload.params.memory

    job = store.get_job(
        name=payload.params.options.job_name,
        version=payload.params.options.job_version,
        db=db,
    )

    if not job:
        error_log.info(f"Received message does not specify a valid job.")
        response.result.errors.append(
            f"The job you specified (name: {payload.params.options.job_name}, version: {payload.params.options.job_version}) does not exist. Is it configured?"
        )
        return fastapi.responses.JSONResponse(
            status_code=fastapi.status.HTTP_200_OK,
            content=response.dict(exclude_none=True),
        )

    # load the supplied message as a dictionary
    received_context = payload.params.message.payload
    if isinstance(received_context, awmodels.PayloadInput):
        received_context = received_context.dict()
    env = liquid.Environment()

    # in a possibly nested context, parse every string value using liquid and render it
    # with the received message payload dictionary.
    context = utils.dict_apply_recursive(
        payload.params.options.context,
        func=lambda x: env.from_string(x).render(received_context),
        filter=lambda x: isinstance(x, str),
    )

    # backwards compatibility: use any context variable supplied in the message under `context`
    if "context" in received_context.keys() and isinstance(
            received_context["context"], dict
    ):
        for k, v in received_context["context"].items():
            # override the key if already exists
            if v is not None:
                context[k] = v

    job_run = store.enqueue_job_run(
        job=job, db=db, context=context, parameters=received_context
    )

    # supply message reponse context param if available
    job_run.tos_response_path = None
    if settings.talend_output_context_param != "":
        job_run.tos_response_path = settings.talend_output_context_param
    if payload.params.options.talend_output_context_param != "":
        job_run.tos_response_path = payload.params.options.talend_output_context_param
    job_run = store.store_job_run(job_run, db)

    p = multiprocessing.Process(
        target=queue.start_job_process, args=[job_run.id, settings], daemon=True
    )
    p.start()

    response.result.logs.append(
        f"Job {job_run.job.name} has been started with id {job_run.id}."
    )
    response.result.memory.jobs.append(job_run.id)

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_202_ACCEPTED,
        content=response.dict(),
    )


@app.post(
    "/tosexecutor/check",
    response_model=awmodels.ResponseCheck,
    response_model_exclude_none=True,
    status_code=fastapi.status.HTTP_200_OK,
)
def check(
        payload: awmodels.RequestCheck,
        db: Session = Depends(store.get_db),
        settings: config.BasicSettings = Depends(get_settings),
):
    """Checks on a running TOS Job or retrieves the output of a completed Job.

    Handle check. In contrast to receive, check cannot start a new TOS Job. The Jobs are differentiated by the
    `job_name` and `job_version` and identified by the `jobs` list of the memory. Any running job not contained
    in the `jobs` list will not be checked / returned.
    """

    access_log.info(f"{datetime.datetime.now().isoformat()} - Check: {payload.dict()}")

    if not utils.is_authorized(payload.params, settings):
        raise errors.UnauthorizedException(
            name="Unauthorized",
            message='Please supply valid credentials using the `api_key` option parameter referring to a stored credential key using Liquid templating: `"api_key": "{% credential credential_name %}"`.',
        )
    response = awmodels.ResponseCheck()
    response.result.memory = payload.params.memory.copy()
    response.result.memory.jobs = []

    job = store.get_job(
        name=payload.params.options.job_name,
        version=payload.params.options.job_version,
        db=db,
    )

    if not job:
        error_log.info(f"Received message does not specify a valid job.")
        response.result.errors.append(
            f"The job you specified (name: {payload.params.options.job_name}, version: {payload.params.options.job_version}) does not exist. Is it configured?"
        )
        return response

    for job_id in payload.params.memory.jobs:
        job_run = store.get_job_run_by_id(job_id, db)
        if job_run and job_run.job == job:
            if job_run.state is intmodels.JobState.queued:
                response.result.logs.append(
                    f"Job {job.name} (V: {job.version}) is queued for running with ID {job_run.id} since {job_run.queued_at.isoformat()}."
                )
                response.result.memory.jobs.append(job_id)

            elif job_run.state is intmodels.JobState.started:
                if not queue.is_running_process(job_run.process_id):
                    response.result.errors.append(
                        f"Job {job.name} (V: {job.version}) running with ID {job_run.id} has been killed from outside of the agent. Consult the logs of the agent to find more information."
                    )
                    job_run.state = intmodels.JobState.broken
                    job_run.retrieved_at = datetime.datetime.now()
                    job_run = store.store_job_run(job_run, db)
                else:
                    response.result.logs.append(
                        f"Job {job.name} (V: {job.version}) is running with ID {job_run.id} since {job_run.started_at.isoformat()}."
                    )
                    response.result.memory.jobs.append(job_id)

            elif job_run.state is intmodels.JobState.completed:
                job_run.retrieved_at = datetime.datetime.now()
                job_run.state = intmodels.JobState.retrieved
                job_run = store.store_job_run(job_run, db)
                run_response = awmodels.ResponseCheck.parse_obj(job_run.response)
                for log in run_response.result.logs:
                    if log != "":
                        response.result.logs.append(log)
                for err in run_response.result.errors:
                    if err != "":
                        response.result.errors.append(err)
                for msg in run_response.result.messages:
                    if msg.returncode == 0:
                        response.result.messages.append(msg)
                        response.result.logs.append(
                            f"Job {job.name} (V: {job.version}) with ID {job_run.id} has completed at {job_run.completed_at.isoformat()}."
                        )
                    else:
                        response.result.errors.append(
                            f"Job {job.name} (V: {job.version}) with ID {job_run.id} has failed at {job_run.completed_at.isoformat()}."
                        )

            elif (
                    job_run.state is intmodels.JobState.retrieved
                    or job_run.state is intmodels.JobState.broken
            ):
                error_log.warning(
                    f"A retrieved or broken job ({job_run.id}) has been checked. Removing it from the memory."
                )

        else:
            error_log.debug(f"Job: {job} | JobID {job_id}")
            error_log.warning(f"Removing orphan job_id {job_id}")

    return response


@app.exception_handler(errors.ActiveWorkflowException)
async def aw_exception_handler(request: Request, exc: errors.ActiveWorkflowException):
    error_log.error(f"{exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_400_BAD_REQUEST,
        content=resp.dict(exclude_none=True),
    )


@app.exception_handler(errors.UnauthorizedException)
async def aw_unauthorized_handler(request: Request, exc: errors.UnauthorizedException):
    error_log.info(f"Unauthorized access - {exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_401_UNAUTHORIZED,
        content=resp.dict(exclude_none=True),
    )


@app.exception_handler(errors.InternalException)
async def aw_internal_exception_handler(
        request: Request, exc: errors.InternalException
):
    error_log.error(f"Internal error - {exc.name}-{exc.job_id}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}-{exc.job_id}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
        content=resp.dict(exclude_none=True),
    )
