# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import subprocess
import os
import os.path
import time


from tos_execute import config
from tos_execute import database
from tos_execute import error_log
from tos_execute import models_internal as intmodel
from tos_execute import store
from tos_execute import utils

from sqlalchemy.orm import Session


def wait_to_start(
    job_run: intmodel.JobRun, db: Session, settings: config.BasicSettings
) -> bool:
    """Will wait in a infinite loop until the job_run_id is the next one allowed to start.

    A job may start when it is the next one in the queue and the number of parallel jobs is
    not greater that the specified in the config.
    """

    hold_run = True
    while hold_run:
        next_job_run = store.get_next_queued_job(job_run.job.id, db)
        if not next_job_run:
            return False

        # is the current job the next queued?
        i_am_up = next_job_run.id is job_run.id

        running_jobs = store.get_job_runs_by_state(intmodel.JobState.started, db)
        # are the running jobs exceeding the parallel jobs count?
        if i_am_up and (len(running_jobs) >= settings.talend_parallel_jobs):
            i_am_up = False
        else:
            # check every running job: is the same as the current job already running?
            for j_r in running_jobs:
                if job_run.job.id is j_r.job.id:
                    i_am_up = False
        if i_am_up:
            next_job_run.started_at = datetime.datetime.now()
            next_job_run.state = intmodel.JobState.started
            job_run = store.store_job_run(next_job_run, db)
            hold_run = False
        else:
            error_log.debug(f"JobRun {job_run.id} sleeps waiting to be scheduled.")
            time.sleep(settings.talend_job_polling_time)
            hold_run = True
    return True


def start_job_process(job_run_id: int, settings: config.BasicSettings) -> None:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session

    # sharing the sessionmaker from database.py does not seem to work here. Using a proper SQLAlchemy engine instead
    engine = create_engine(settings.db_uri)
    db = Session(bind=engine, autoflush=False, autocommit=False)

    job_run = store.get_job_run_by_id(job_run_id, db)

    if not job_run:
        error_log.critical(
            f"Cannot find JobRun {job_run_id} in the database. Shutting down Process."
        )
        return

    job_run.process_id = int(os.getpid())
    error_log.debug(f"Starting Process ({job_run.process_id}) for JobRun {job_run.id}.")

    # do all necessary preperations for a job execution
    job_run, cmd = utils.prepare_job_execution(job_run)

    store.store_job_run(job_run, db)
    # wait to be scheduled
    if not wait_to_start(job_run, db, settings):
        error_log.critical(
            f"Some internal error happened: getting the next job returned None in a Process. JobRun ID: {job_run.id}"
        )
        return

    error_log.debug(
        f"Running job {job_run.id} ({job_run.job.name}, V: {job_run.job.version}) with command: {cmd}"
    )
    my_process = subprocess.run(
        " ".join(cmd), capture_output=True, encoding="utf-8", shell=True
    )
    error_log.debug(
        f"Job {job_run.id} ({job_run.job.name}, V: {job_run.job.version}) has completed with code {my_process.returncode}."
    )

    # do all necessary post-run processing
    job_run = utils.post_process_job_run(job_run, my_process)

    error_log.debug(
        f"Post-processing of JobRun {job_run.id} done. Response:\n{job_run.response}"
    )

    job_run = store.store_job_run(job_run, db)
    error_log.info(
        f"Job {job_run.job.name} (V: {job_run.job.version}) ran in process {job_run.process_id} with id {job_run.id} and finished at {job_run.completed_at.isoformat()}."
    )

    # close the db session (should happen automatically though)
    db.close()


def is_running_process(process_id: int) -> bool:
    """Returns whether a process_id is found under /proc, indicating this process is still running.

    This method only works on Linux!
    It can be used to find out whether a process has been killed by the OS due to memory constraints.

    Parameters
    ----------
    process_id: int
        The process ID of the process that will be checked.

    Results
    -------
    bool
        True if the process is running, False otherwise.
    """

    if os.path.exists(f"/proc/{process_id}"):
        return True
    return False
