# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime

from typing import Any, Dict, List, Optional

from sqlalchemy import and_, update
from sqlalchemy.orm import Session

from tos_execute import database
from tos_execute import models_internal as intmodel


def get_db():
    db = database.SessionLocal()
    try:
        yield db
    finally:
        db.close()


def get_jobs_by_availability(avail: bool, db: Session) -> List[intmodel.Job]:
    return db.query(intmodel.Job).filter(intmodel.Job.available == avail).all()


def get_job(name: str, version: str, db: Session) -> Optional[intmodel.Job]:
    return (
        db.query(intmodel.Job)
        .filter(and_(intmodel.Job.name == name, intmodel.Job.version == version))
        .first()
    )


def set_all_jobs_inactive(db: Session) -> None:
    db.execute(
        update(intmodel.Job)
        .where(intmodel.Job.available == True)
        .values(available=False)
    )


def store_job(job: intmodel.Job, db: Session) -> intmodel.Job:
    db.add(job)
    db.commit()
    db.refresh(job)
    return job


def enqueue_job_run(
    job: intmodel.Job, db: Session, context: Dict[str, Any], parameters: Dict[str, Any]
) -> intmodel.JobRun:
    job_run = intmodel.JobRun(
        job=job,
        state=intmodel.JobState.queued,
        queued_at=datetime.datetime.now(),
        context=context,
        job_parameters=parameters,
    )
    db.add(job_run)
    db.commit()
    db.refresh(job_run)
    return job_run


def get_next_queued_job(job_id: int, db: Session) -> Optional[intmodel.JobRun]:
    return (
        db.query(intmodel.JobRun)
        .filter(
            intmodel.JobRun.state == intmodel.JobState.queued, intmodel.Job.id == job_id
        )
        .order_by(intmodel.JobRun.queued_at.asc())
        .first()
    )


def store_job_run(job_run: intmodel.JobRun, db: Session) -> intmodel.JobRun:
    db.add(job_run)
    db.commit()
    db.refresh(job_run)
    return job_run


def get_job_run_by_id(job_id: int, db: Session) -> Optional[intmodel.JobRun]:
    return db.query(intmodel.JobRun).filter(intmodel.JobRun.id == job_id).first()


def get_job_runs_by(
    db: Session, job_id: int = None, job_run_state: intmodel.JobState = None
) -> List[intmodel.JobRun]:
    if job_id and job_run_state:
        return (
            db.query(intmodel.JobRun)
            .filter(intmodel.JobRun.state == job_run_state, intmodel.Job.id == job_id)
            .all()
        )

    if job_id:
        return db.query(intmodel.JobRun).filter(intmodel.Job.id == job_id).all()

    if job_run_state:
        return (
            db.query(intmodel.JobRun)
            .filter(intmodel.JobRun.state == job_run_state)
            .all()
        )

    return []


def get_job_runs_by_state(
    job_run_state: intmodel.JobState, db: Session
) -> List[intmodel.JobRun]:
    return (
        db.query(intmodel.JobRun).filter(intmodel.JobRun.state == job_run_state).all()
    )
