# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import random
import string
import time

import pytest
from fastapi.testclient import TestClient

from tos_execute import config
from tos_execute import main

testclient = TestClient(main.app)


def test_joblist():
    """
    Tests if a list of all jobs is available from the agent.
    """

    response = testclient.get("/joblist")
    assert 200 <= response.status_code <= 299

    response_json = response.json()
    assert type(response_json) == list
    assert len(response_json) == 1
    assert type(response_json[0]) == dict
    assert "name" in response_json[0].keys()
    assert "version" in response_json[0].keys()


@pytest.fixture(scope="function")
def api_key():
    """
    Fixture to enable api key authorization for this agent and disable it after supplying the api key
    """
    api_key = "myCoolApiKeyNobodyKnows"

    main.app.dependency_overrides[main.get_settings] = lambda: config.BasicSettings(
        application_api_keys=[api_key, "other_apikey"]
    )

    yield api_key

    main.app.dependency_overrides[main.get_settings] = lambda: config.BasicSettings()


@pytest.fixture(scope="function")
def store_directory():
    """
    Fixture to create a random filename under tmp and delete any file that may be created after a test has run.
    """
    store_directory = (
            "/tmp/" + "".join(random.choices(string.ascii_letters, k=8)) + "_out.csv"
    )
    yield store_directory
    if os.path.exists(store_directory):
        os.unlink(store_directory)


def test_apikey_authorization(api_key: str):
    """
    Test if api key authorization is enforced if configured.
    """
    payload = {
        "method": "check",
        "params": {
            "message": None,
            "options": {"job_name": "testJob", "job_version": "0.2"},
            "memory": {},
            "credentials": [],
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert response.status_code == 401

    payload["params"]["credentials"].append({"name": "invalid", "value": "value"})
    response = testclient.post("/tosexecutor", json=payload)
    assert response.status_code == 401

    payload["params"]["options"]["api_key"] = "{% credential marcels_apikey %}"
    response = testclient.post("/tosexecutor", json=payload)
    # we expect 400 error in this case as the credential reference to a
    # missing value is raised in `util.replace_credentials()`
    assert response.status_code == 400

    payload["params"]["credentials"][0] = {
        "name": "marcels_apikey",
        "value": api_key,
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299


def test_receive_message():
    """
    Test the compliance to the ActiveWorkflow register message.
    See https://github.com/automaticmode/active_workflow/wiki/Remote-Agent-API#the-register-method
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {"method": "register", "params": {}}
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299

    response_json = response.json()
    assert response_json
    assert type(response_json) == dict
    assert "result" in response_json.keys()
    for key in ["name", "display_name", "description", "default_options"]:
        assert key in response_json["result"]

    assert response_json["result"][
               "name"
           ] == config.BasicSettings().application_name.replace(" ", "")
    for key in ["name", "display_name", "description"]:
        assert type(response_json["result"][key]) == str
    assert type(response_json["result"]["default_options"]) == dict


def test_empty_check():
    """
    Test checking an empty TOS Agent instance.
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {
        "method": "check",
        "params": {
            "message": None,
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
            },
            "memory": {},
            "credentials": [],
        },
    }
    response = testclient.post("/tosexecutor", json=payload)

    assert 200 <= response.status_code <= 299
    response_json = response.json()
    assert response_json
    assert type(response_json) == dict
    assert "result" in response_json.keys()
    for key in ["errors", "logs", "memory", "messages"]:
        assert key in response_json["result"]

    for key in ["errors", "logs", "messages"]:
        assert type(response_json["result"][key]) == list
        assert len(response_json["result"][key]) == 0

    assert type(response_json["result"]["memory"]) == dict


def test_performJob_emptyPayload(store_directory):
    """
    Test to start a TOS job with an empty payload (i.e. use default settings on the agent) on the TOS agent.
    Also, this test method will check the compliance to ActiveWorkflows Receive message.
    See https://github.com/automaticmode/active_workflow/wiki/Remote-Agent-API#the-receive-method
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {
        "method": "receive",
        "params": {
            "message": {"payload": {}},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {},
            "credentials": [],
        },
    }

    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299

    response_json = response.json()
    assert response_json
    assert type(response_json) == dict
    assert "result" in response_json.keys()
    for key in ["errors", "logs", "memory", "messages"]:
        assert key in response_json["result"]

    assert type(response_json["result"]["memory"]) == dict
    assert "jobs" in response_json["result"]["memory"]
    assert type(response_json["result"]["memory"]["jobs"]) == list
    assert len(response_json["result"]["memory"]["jobs"]) == 1

    job_id = response_json["result"]["memory"]["jobs"][0]
    assert job_id
    assert type(job_id) == int

    # giving the agent time to execute the TOS Job
    time.sleep(5)

    payload = {
        "method": "check",
        "params": {
            "message": {},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {"jobs": [job_id]},
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299

    # note: will not recheck structure of check message, this is done in another unit test
    # awesome oneliner to read the lines of a file by christian.schmidt2@med.uni-goettingen.de
    assert sum(1 for line in open(store_directory)) == 7


def test_performJob_responseMessage(store_directory):
    """
    Test to run a TOS job and retrieve a response message.
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {
        "method": "receive",
        "params": {
            "message": {"payload": {}},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "talend_output_context_param": "aw_message_filepath",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {},
            "credentials": [],
        },
    }

    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_id = response_json["result"]["memory"]["jobs"][0]

    # giving the agent time to execute the TOS Job
    time.sleep(5)

    payload = {
        "method": "check",
        "params": {
            "message": {},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "talend_output_context_param": "aw_message_filepath",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {"jobs": [job_id]},
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()

    assert len(response_json["result"]["messages"]) == 1
    assert type(response_json["result"]["messages"][0]) == dict
    assert "response" in response_json["result"]["messages"][0]
    assert type(response_json["result"]["messages"][0]["response"]) == list
    # check if the output of the Talendjob is correct
    tos_response_msg = response_json["result"]["messages"][0]["response"]
    assert len(tos_response_msg) == 7


def test_performJob_payloadOverride(store_directory):
    """
    Test the override functionality of supplying a payload.
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {
        "method": "receive",
        "params": {
            "message": {"payload": {"context": {"rows": 19}}},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {},
            "credentials": [],
        },
    }

    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_id = response_json["result"]["memory"]["jobs"][0]

    # giving the agent time to execute the TOS Job
    time.sleep(5)

    payload = {
        "method": "check",
        "params": {
            "message": {},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {"jobs": [job_id]},
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299

    # note: will not recheck structure of check message, this is done in another unit test
    # awesome oneliner to read the lines of a file by christian.schmidt2@med.uni-goettingen.de
    assert sum(1 for line in open(store_directory)) == 19


def test_performJob_liquidOverride(store_directory):
    """
    Test the liquid templating override functionality of supplying a payload.
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {
        "method": "receive",
        "params": {
            "message": {"payload": {"otherthing": {"i_need": 12}}},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {
                    "rows": "{{ otherthing.i_need }}",
                    "storeDirectory": store_directory,
                },
            },
            "memory": {},
            "credentials": [],
        },
    }

    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_id = response_json["result"]["memory"]["jobs"][0]

    # giving the agent time to execute the TOS Job
    time.sleep(5)

    payload = {
        "method": "check",
        "params": {
            "message": {},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {
                    "rows": "{{ otherthing.i_need }}",
                    "storeDirectory": store_directory,
                },
            },
            "memory": {"jobs": [job_id]},
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299

    # note: will not recheck structure of check message, this is done in another unit test
    # awesome oneliner to read the lines of a file by christian.schmidt2@med.uni-goettingen.de
    assert sum(1 for line in open(store_directory)) == 12


def test_performJob_queueTest(store_directory):
    """
    Test if incoming job executions are queued properly.
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")
    job_ids = []

    # Job no 1
    payload = {
        "method": "receive",
        "params": {
            "message": {"payload": {}},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {},
            "credentials": [],
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_ids.append(response_json["result"]["memory"]["jobs"][0])

    # Job no 2
    time.sleep(3)
    payload["params"]["options"]["context"]["rows"] = 19
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_ids.append(response_json["result"]["memory"]["jobs"][0])

    # Job no 3
    time.sleep(3)
    payload["params"]["options"]["context"]["rows"] = 2
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_ids.append(response_json["result"]["memory"]["jobs"][0])

    payload = {
        "method": "check",
        "params": {
            "message": {},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory},
            },
            "memory": {"jobs": job_ids},
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    # due to job polling time, we expect not all Jobs have run at this point,
    # yielding at least one job id left in the job_ids memory
    job_ids = response_json["result"]["memory"]["jobs"]
    assert len(job_ids) >= 1

    # wait configured time to let the jobs finish
    time.sleep(config.BasicSettings().talend_job_polling_time * 2)

    # awesome oneliner to read the lines of a file by christian.schmidt2@med.uni-goettingen.de
    assert sum(1 for line in open(store_directory)) == 2


def test_performJob_errorResponse(store_directory):
    """
    Let the test TOS job fail and check if this fact is responded accordingly (e.g. no message is generated).
    """
    # make sure the jobs are loaded
    testclient.get("/joblist")

    payload = {
        "method": "receive",
        "params": {
            "message": {"payload": {}},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory, "fail": True},
            },
            "memory": {},
            "credentials": [],
        },
    }

    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    job_id = response_json["result"]["memory"]["jobs"][0]

    # giving the agent time to execute the TOS Job
    time.sleep(5)

    payload = {
        "method": "check",
        "params": {
            "message": {},
            "options": {
                "job_name": "testJob",
                "job_version": "0.2",
                "context": {"rows": 7, "storeDirectory": store_directory, "fail": True},
            },
            "memory": {"jobs": [job_id]},
        },
    }
    response = testclient.post("/tosexecutor", json=payload)
    assert 200 <= response.status_code <= 299
    response_json = response.json()
    assert len(response_json["result"]["errors"]) >= 1
    assert len(response_json["result"]["messages"]) == 0
