# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import json
import os
import re as regex
import shlex
import subprocess
import tempfile
from typing import Tuple, Dict, Any, List, Union, Optional

from tos_execute import config
from tos_execute import error_log
from tos_execute import models_activeworkflow as awmodels
from tos_execute import models_internal as intmodels
from tos_execute.errors import ActiveWorkflowException


def prepare_job_execution(
        job_run: intmodels.JobRun,
) -> Tuple[intmodels.JobRun, List[str]]:
    """Prepare the execution of a TOS job based on the JobRun and the context

    This functional method will prepare all necessary steps prior to a job execution and return
    updated versions of a JobRun, the context dictionary as well as the run-command list.

    Parameters
    ----------
    job_run: intmodels.JobRun
        The JobRun instance that is stored as queued in the database.

    Returns
    -------
    intmodels.JobRun
        The JobRun instance with possible modifications according to the pre-process workflow.
    List[str]
        The job start shell command as a List
    """

    # Check if TOS output is set, create file path for it if necessary
    if job_run.tos_response_path is not None:
        talend_output_path = os.path.join(
            tempfile.gettempdir(),
            f"{job_run.job.name.lower()}_{job_run.job.version.lower()}_{job_run.id}.json",
        )
        job_run.context[job_run.tos_response_path] = talend_output_path
        job_run.tos_response_path = talend_output_path

    context_params = [
        f"--context_param={shlex.quote(param)}={shlex.quote(str(val))}"
        for param, val in job_run.context.items()
    ]

    os.chmod(job_run.job.path, 0o755)  # make the exec sript executable
    cmd = [f"{job_run.job.path}"] + context_params

    return (job_run, cmd)


def post_process_job_run(
        job_run: intmodels.JobRun,
        process: subprocess.CompletedProcess,
) -> intmodels.JobRun:
    """Conduct all necessary steps to post-process a run job.

    This functional method takes all steps necessary to post process a run job based
    on the JobRun and the CompletedProcess instance. It will set some stats
    as well as the possible response message.

    Paramters
    ---------
    job_run: intmodels.JobRun
        The JobRun instance in started state.
    process: subprocess.CompletedProcess
        The CompletedProcess instance in which the TOS job has run.

    Returns
    -------
    intmodels.JobRun
        The JobRun instance with possible modifications according to the post-process workflow.
    """

    # set stats
    job_run.completed_at = datetime.datetime.now()
    job_run.state = intmodels.JobState.completed

    # set response
    response = awmodels.ResponseCheck()
    response.result.messages.append(
        awmodels.MessageOutput(
            job_run_id=job_run.id,
            job_id=job_run.job.id,
            job_name=job_run.job.name,
            job_version=job_run.job.version,
            job_parameters=job_run.job_parameters,
            start_time=job_run.started_at.isoformat(),
            end_time=job_run.completed_at.isoformat(),
        )
    )

    # set stdout / stderr
    if len(process.stdout) >= 1:
        response.result.logs = [process.stdout]
    if len(process.stderr) >= 1:
        # TODO: ugly ugly ugly!!
        cleaned = remove_slf4j_errors(process.stderr)
        if cleaned:
            response.result.errors = [cleaned]

    # set / calculate the job run stats
    response.result.messages[0].completed = True
    response.result.messages[0].returncode = process.returncode
    response.result.messages[0].runtime = (
            job_run.completed_at - job_run.started_at
    ).total_seconds()

    # set / retrieve the job run response
    if job_run.tos_response_path is not None:
        response.result.messages[0].response = retrieve_output_file(
            job_run.tos_response_path
        )
    else:
        response.result.messages[0].response = {}
    # hacking dirty: write every date as a JSON string and re-parse it as a dictionary for response.
    job_run.response = json.loads(response.json(exclude_none=True))

    return job_run


def retrieve_output_file(path: str) -> Union[Dict[Any, Any], List[Any]]:
    """Retrieve a TOS job output from a path.

    Retrieves a reponse that has been written by the TOS job to path. Return
    this file as a pythonized JSON type (Dict or List).

    Parameters
    ----------
    path: str
        The full path to the file that has been written by a TOS job.

    Results
    -------
    Union[Dict[Any, Any], List[Any]]
        A pythonized JSON, read from the message written be a TOS job. Should
        be either Dict[Any, Any] (JSON Object) or a List[Any] (JSON Array)
    """
    response = {}
    if path != "" and os.path.exists(path) and os.path.isfile(path):
        try:
            with open(path, "r") as response_file:
                response = json.load(response_file)
        except:
            # Cannot read? no response then
            error_log.error(f"Could not read the output file: {path}")
            response = {}
    else:
        error_log.warning(f"Expected a file at path {path}, but it is invalid.")
    return response


def remove_slf4j_errors(stderr: Optional[str]) -> Optional[str]:
    """Dirty hack method: remove all SLF4J errors from a String.

    Each line in the String will be checked for "SLF4J". If a line
    starts with this part, remove this line from the stderr input
    str. Return all other lines.

    Parameters
    ----------
    stderr: Optional[str]
        Any string. Should be a multiline string that is the stderr output
        from a CompletedProcess.

    Results
    -------
    Optional[str]
        The input string where every line starting with "SLF4J:" has been removed.
    """
    if not stderr:
        None
    out_str = ""
    for line in stderr.splitlines(keepends=True):
        if line.startswith("SLF4J:"):
            pass
        else:
            out_str += line
    return out_str


def is_authorized(
        params: Union[
            awmodels.ParamsRegister, awmodels.ParamsReceive, awmodels.ParamsCheck
        ],
        settings: config.BasicSettings,
) -> bool:
    """Check if a params section is authorized to use this TOS agent instance.

    This method will check 1) if credentials are set in the configuration of the agent and 2)
    whether the supplied params are authorized to use the agent. If no credentials are set
    (which is the default), the method will return True.

    Parameters
    ----------
    params: Union[awmodels.ParamsRegister, awmodels.ParamsReceive, awmodels.ParamsCheck]
        Any params dictionary posted by the requester JSON

    Results
    -------
    bool
        True if parameters hold valid credentials, authorizing any request. False otherwise
    """

    if len(settings.application_api_keys) == 0:
        # If there are no credentials set, no credentials are needed.
        return True

    if params.options.api_key in settings.application_api_keys:
        return True

    return False


def dict_apply_recursive(dct: Dict[str, Any], func, filter) -> Dict[str, Any]:
    """Applies `func` to every item in `dct`, where `filter` returns True.

    This method will apply itself (or the list variant) recursively to any item in
    `dct` of type Dict or List.
    """
    result_dct = {}
    for k, v in dct.items():
        if isinstance(v, dict):
            result_dct[k] = dict_apply_recursive(v, func, filter)
        elif isinstance(v, list):
            result_dct[k] = list_apply_recursive(v, func, filter)
        elif filter(v):
            result_dct[k] = func(v)
        else:
            result_dct[k] = v
    return result_dct


def list_apply_recursive(lst: List[Any], func, filter) -> List[Any]:
    """Applies `func` to every item in `lst` where `filter` returns True.

    This method will apply itself (or the dict variant) recursively to any item in `lst` of type
    Dict or List.
    """
    result_lst = []
    for v in lst:
        if isinstance(v, dict):
            result_lst.append(dict_apply_recursive(v, func, filter))
        elif isinstance(v, list):
            result_lst.append(list_apply_recursive(v, func, filter))
        elif filter(v):
            result_lst.append(func(v))
        else:
            result_lst.append(v)
    return result_lst


def replace_credentials(data: Dict, credentials: List[awmodels.CredentialsCommon]):
    """
    Replace occurrences of AW credential syntax in the items of a Dict

    Searches for ActiveWorkflow credential references in Dict items' values, i.e. `{% credential ref_id %}` and
    replaces the entire string with the value of `ref_id` if that is present as a key in the credentials Dict.

    Parameters
    ----------
    data: Dict
        Any dictionary that possibly contains ActiveWorkflow credential references
    credentials: List[awmodels.CredentialsCommon]
        A List of CredentialsCommon objects

    Results
    -------
    data: Dict
        The input Dict with replaced values (if applicable)
    """
    for key, value in data.items():
        # Enter recursion if value itself is a dictionary
        if type(value) is dict:
            data[key] = replace_credentials(value, credentials)
        # determine whether value contains an AW credential reference
        # and look up the reference in the given list of credentials

        if type(value) is str:
            # Regular expression tests whether a reference to ActiveWorkflow credential is present,
            # designated by double-curly-brackets: {% credential credential_name %}
            regex_str = "({% credential )(.[^%}]*)( %})"
            for matches in regex.finditer(regex_str, value):
                if matches:
                    # [0] is the whole string if the regex is matched, [1] is the first group, [2] the second group...
                    # We need the 2nd group "(.*)" from the regex, which would be the name of the credential
                    reference = matches[2].strip()
                    replaced = False
                    for c in credentials:
                        if c.name == reference:
                            value = value.replace(matches[0], c.value)
                            data[key] = value
                            replaced = True
                    if not replaced:
                        raise ActiveWorkflowException(name="Missing Credential",
                                                      message=f"Credential '{reference}' is not supplied.")
    return data
